# StellarGraph demos


[StellarGraph](https://github.com/stellargraph/stellargraph) provides numerous algorithms for graph machine learning. This folder contains demos of all of them to explain how they work and how to use them as part of a TensorFlow Keras data science workflow.

The demo notebooks can be run without any installation of Python by using Binder or Google Colab - these both provide a cloud-based notebook environment.  The whole set of demos can be opened in Binder here: [![Open in Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/stellargraph/stellargraph/master?urlpath=lab/tree/demos/README.md) or you can click the Binder and Colab badges within each notebook.

## Find algorithms for a task

- Introduction to StellarGraph and its graph machine learning workflow (with TensorFlow and Keras): [**GCN on Cora**](node-classification/gcn/gcn-cora-node-classification-example.ipynb)
- Predicting attributes, such as classifying as a class or label, or regressing to calculate a continuous number:
  - For nodes/vertices/entities: [**node classification**](./node-classification/README.md)
  - For edges/links/connections: [**link prediction**](./link-prediction/README.md) (includes knowledge graph completion)
  - For graphs/networks: [**graph classification**](./graph-classification/README.md)
  - Adjusting predictions scores to be probabilities (for any model): [**calibration**](./calibration/README.md)
  - Interpreting/introspecting models, for node classification: [**interpretability**](./interpretability/README.md)
- Representation learning or computing embedding vectors (including unsupervised tasks):
  - For nodes/vertices/entities and edges/links/connections: [**embeddings**](./embeddings/README.md) (and [**community detection**](./community_detection/README.md) for using these to cluster nodes)
- Ensembling models to reduce prediction variance: [**ensembles**](./ensembles/README.md)
- Loading data into a `StellarGraph` object, with Pandas, Neo4j or NetworkX: [**basics**](./basics/README.md)
- Experimental: running GraphSAGE on data stored in Neo4j: [**neo4j connector**](./connector/neo4j/README.md)

## Find a demo for an algorithm

(Hover for explanation of abbreviations.)

<!-- DEMO TABLE MARKER -->
<!-- autogenerated by scripts/demo_table.py, edit that file instead of this location -->
<table>
  <tr>
    <th>Algorithm</th>
    <th><span title='Heterogeneous graphs'>Heter.</span></th>
    <th><span title='Directed'>Dir.</span></th>
    <th><span title='Edge weights'>EW</span></th>
    <th><span title='Time-varying or temporal graphs'>T</span></th>
    <th><span title='Node features'>NF</span></th>
    <th><span title='Node classification'><a href='node-classification/README.md'>NC</a></span></th>
    <th><span title='Link prediction'><a href='link-prediction/README.md'>LP</a></span></th>
    <th><span title='Representation learning'><a href='embeddings/README.md'>RL</a></span></th>
    <th><span title='Inductive'>Ind.</span></th>
    <th><span title='Graph classification'><a href='graph-classification/README.md'>GC</a></span></th>
  </tr>
  <tr>
    <td><span title='Graph Convolutional Network'>GCN</span></td>
    <td>see RGCN</td>
    <td></td>
    <td></td>
    <td></td>
    <td>yes</td>
    <td><a href='node-classification/gcn/gcn-cora-node-classification-example.ipynb'>demo</a></td>
    <td><a href='link-prediction/gcn/cora-gcn-links-example.ipynb'>demo</a></td>
    <td>
      <span title='UnsupervisedSampler, using link prediction'>US</span>
      <span title='DeepGraphInfomax, using mutual information'><a href='embeddings/deep-graph-infomax-cora.ipynb'>DGI</a></span>
    </td>
    <td>see Cluster-GCN</td>
    <td><a href='graph-classification/supervised-graph-classification.ipynb'>demo</a></td>
  </tr>
  <tr>
    <td>Cluster-GCN</td>
    <td></td>
    <td></td>
    <td></td>
    <td></td>
    <td>yes</td>
    <td><a href='node-classification/cluster-gcn/cluster-gcn-node-classification.ipynb'>demo</a></td>
    <td>yes</td>
    <td></td>
    <td>yes</td>
    <td></td>
  </tr>
  <tr>
    <td><span title='Relational GCN'>RGCN</span></td>
    <td><span title='Multiple edges types and one node type'>yes, edges</span></td>
    <td></td>
    <td></td>
    <td></td>
    <td>yes</td>
    <td><a href='node-classification/rgcn/rgcn-aifb-node-classification-example.ipynb'>demo</a></td>
    <td>yes</td>
    <td></td>
    <td></td>
    <td></td>
  </tr>
  <tr>
    <td><span title='Graph ATtention Network'>GAT</span></td>
    <td></td>
    <td></td>
    <td></td>
    <td></td>
    <td>yes</td>
    <td><a href='node-classification/gat/gat-cora-node-classification-example.ipynb'>demo</a></td>
    <td>yes</td>
    <td>
      <span title='UnsupervisedSampler, using link prediction'>US</span>
      <span title='DeepGraphInfomax, using mutual information'><a href='embeddings/deep-graph-infomax-cora.ipynb'>DGI</a></span>
    </td>
    <td></td>
    <td></td>
  </tr>
  <tr>
    <td><span title='Simplified Graph Convolution'>SGC</span></td>
    <td></td>
    <td></td>
    <td></td>
    <td></td>
    <td>yes</td>
    <td><a href='node-classification/sgc/sgc-node-classification-example.ipynb'>demo</a></td>
    <td>yes</td>
    <td></td>
    <td></td>
    <td></td>
  </tr>
  <tr>
    <td><span title='Personalized Propagation of Neural Predictions'>PPNP</span></td>
    <td></td>
    <td></td>
    <td></td>
    <td></td>
    <td>yes</td>
    <td><a href='node-classification/ppnp/ppnp-cora-node-classification-example.ipynb'>demo</a></td>
    <td>yes</td>
    <td>
      <span title='UnsupervisedSampler, using link prediction'>US</span>
      <span title='DeepGraphInfomax, using mutual information'>DGI</span>
    </td>
    <td></td>
    <td></td>
  </tr>
  <tr>
    <td><span title='Approximate PPNP'>APPNP</span></td>
    <td></td>
    <td></td>
    <td></td>
    <td></td>
    <td>yes</td>
    <td><a href='node-classification/ppnp/ppnp-cora-node-classification-example.ipynb'>demo</a></td>
    <td>yes</td>
    <td>
      <span title='UnsupervisedSampler, using link prediction'>US</span>
      <span title='DeepGraphInfomax, using mutual information'><a href='embeddings/deep-graph-infomax-cora.ipynb'>DGI</a></span>
    </td>
    <td></td>
    <td></td>
  </tr>
  <tr>
    <td>GraphWave</td>
    <td></td>
    <td></td>
    <td></td>
    <td></td>
    <td></td>
    <td><span title='As a downstream task by training a classifier on reprentation/embedding vectors'>via RL</span></td>
    <td><span title='As a downstream task by training a classifier on reprentation/embedding vectors'>via RL</span></td>
    <td><a href='embeddings/graphwave-barbell.ipynb'>demo</a></td>
    <td></td>
    <td></td>
  </tr>
  <tr>
    <td>Attri2Vec</td>
    <td></td>
    <td></td>
    <td></td>
    <td></td>
    <td>yes</td>
    <td><a href='node-classification/attri2vec/attri2vec-citeseer-node-classification-example.ipynb'>demo</a></td>
    <td><a href='link-prediction/attri2vec/stellargraph-attri2vec-DBLP.ipynb'>demo</a></td>
    <td><a href='embeddings/stellargraph-attri2vec-citeseer.ipynb'>demo</a></td>
    <td></td>
    <td></td>
  </tr>
  <tr>
    <td>GraphSAGE</td>
    <td>see HinSAGE</td>
    <td><a href='node-classification/graphsage/directed-graphsage-on-cora-example.ipynb'>demo</a></td>
    <td></td>
    <td></td>
    <td>yes</td>
    <td><a href='node-classification/graphsage/graphsage-cora-node-classification-example.ipynb'>demo</a></td>
    <td><a href='link-prediction/graphsage/cora-links-example.ipynb'>demo</a></td>
    <td>
      <span title='UnsupervisedSampler, using link prediction'><a href='embeddings/embeddings-unsupervised-graphsage-cora.ipynb'>US</a></span>
      <span title='DeepGraphInfomax, using mutual information'><a href='embeddings/deep-graph-infomax-cora.ipynb'>DGI</a></span>
    </td>
    <td><a href='node-classification/graphsage/graphsage-pubmed-inductive-node-classification-example.ipynb'>demo</a></td>
    <td></td>
  </tr>
  <tr>
    <td>HinSAGE</td>
    <td>yes</td>
    <td></td>
    <td></td>
    <td></td>
    <td>yes</td>
    <td>yes</td>
    <td><a href='link-prediction/hinsage/movielens-recommender.ipynb'>demo</a></td>
    <td><span title='DeepGraphInfomax, using mutual information'><a href='embeddings/deep-graph-infomax-cora.ipynb'>DGI</a></span></td>
    <td>yes</td>
    <td></td>
  </tr>
  <tr>
    <td>Node2Vec</td>
    <td></td>
    <td></td>
    <td><a href='node-classification/node2vec/stellargraph-node2vec-weighted-random-walks.ipynb'>demo</a></td>
    <td></td>
    <td></td>
    <td><span title='As a downstream task by training a classifier on reprentation/embedding vectors'><a href='node-classification/node2vec/stellargraph-node2vec-node-classification.ipynb'>via RL</a></span></td>
    <td><span title='As a downstream task by training a classifier on reprentation/embedding vectors'><a href='link-prediction/random-walks/cora-lp-demo.ipynb'>via RL</a></span></td>
    <td><a href='embeddings/stellargraph-node2vec.ipynb'>demo</a></td>
    <td></td>
    <td></td>
  </tr>
  <tr>
    <td>Metapath2Vec</td>
    <td>yes</td>
    <td></td>
    <td></td>
    <td></td>
    <td></td>
    <td><span title='As a downstream task by training a classifier on reprentation/embedding vectors'>via RL</span></td>
    <td><span title='As a downstream task by training a classifier on reprentation/embedding vectors'>via RL</span></td>
    <td><a href='embeddings/stellargraph-metapath2vec.ipynb'>demo</a></td>
    <td></td>
    <td></td>
  </tr>
  <tr>
    <td><span title='Continuous-Time Dynamic Network Embeddings'>CTDNE</span></td>
    <td></td>
    <td></td>
    <td></td>
    <td>yes</td>
    <td></td>
    <td><span title='As a downstream task by training a classifier on reprentation/embedding vectors'>via RL</span></td>
    <td><span title='As a downstream task by training a classifier on reprentation/embedding vectors'><a href='link-prediction/random-walks/ctdne-link-prediction.ipynb'>via RL</a></span></td>
    <td>yes</td>
    <td></td>
    <td></td>
  </tr>
  <tr>
    <td>Watch Your Step</td>
    <td></td>
    <td></td>
    <td></td>
    <td></td>
    <td></td>
    <td><span title='As a downstream task by training a classifier on reprentation/embedding vectors'><a href='embeddings/watch-your-step-cora-demo.ipynb'>via RL</a></span></td>
    <td><span title='As a downstream task by training a classifier on reprentation/embedding vectors'>via RL</span></td>
    <td><a href='embeddings/watch-your-step-cora-demo.ipynb'>demo</a></td>
    <td></td>
    <td></td>
  </tr>
  <tr>
    <td>ComplEx</td>
    <td><span title='Multiple edges types and one node type'>yes, edges</span></td>
    <td>yes</td>
    <td></td>
    <td></td>
    <td></td>
    <td><span title='As a downstream task by training a classifier on reprentation/embedding vectors'>via RL</span></td>
    <td><a href='link-prediction/knowledge-graphs/complex.ipynb'>demo</a></td>
    <td>yes</td>
    <td></td>
    <td></td>
  </tr>
  <tr>
    <td>DistMult</td>
    <td><span title='Multiple edges types and one node type'>yes, edges</span></td>
    <td>yes</td>
    <td></td>
    <td></td>
    <td></td>
    <td><span title='As a downstream task by training a classifier on reprentation/embedding vectors'>via RL</span></td>
    <td><a href='link-prediction/knowledge-graphs/distmult.ipynb'>demo</a></td>
    <td>yes</td>
    <td></td>
    <td></td>
  </tr>
</table>
<!-- DEMO TABLE MARKER -->

See [the root README](../README.md) or each algorithm's documentation for the relevant citation(s).
